// acd_scal.cc
// Author: Mario J. Bencomo
// last modified: 12/12/16

#include "acd_defn.hh"
#include "acd.hh"
#include "MPS_includes.hh"
#include "MPS_Space_Examples.hh"
#include "MPS_to_RHS.hh"

using RVL::valparse;
using RVL::AssignFilename;
using RVL::OperatorEvaluation;
using TSOpt::Scal_MPS_Space;
using TSOpt::MPS_to_RHS;
using TSOpt::add_to_pars;

const char *sdoc[] = {
  " ============================================================================",
  " acd_scal.x ",
  " ============================================================================",
  " Authors: Mario J. Bencomo",
  " ",
  " Acoustic constant density centered difference modeling for single scalar",
  " multipole sources.", 
  " ",
  " PDE:",
  "   (d/dt)^2 p - c^2 nabla^2 p = f(x,t)",
  " Source term:",
  "   2-D, f(x,t) = f_0(t) (d/dx_0)^{s_0} (d/dx_1)^{s_1} delta(x-x^*)",
  "   3-D, f(x,t) = f_0(t) (d/dx_0)^{s_0} (d/dx_1)^{s_1} (d/dx_2)^{s_2} delta(x-x^*)",
  " ",
  " Typical parameter list. May be copied, edited, and used for input: either",
  " include parameters on command line (for example in Flow), or place",
  " in file <foo> and include \"par=<foo>\" on command line. Any parameter",
  " included explicitly in command line overrides parameter with same key",
  " in par file.",
  " ",
  "  Invoke single threaded execution by ",
  " \"acd_scal.x [parameters] [standalone install]\"",
  " ",
  " or multi-threaded execution using interactive or batch MPI (for which",
  " code must be compiled with MPI enabled).",
  " ",
  " Given values are defaults; non-optional values indicated by corner brackets.",
  " ",
  " --------------------------- begin parameters ---------------------------",
  " FD info:",
  " ",
  "          order = 2           spatial half-order",
  "            cfl = 0.75        proportion of max dt/dx",
  "           cmin = 1.0         min permitted velocity (m/ms) - sanity check",
  "           cmax = 4.5         max permitted velocity (m/ms) - used in dt comp",
  " ",
  " ------------------------------------------------------------------------",
  " Source info:",
  " ",
  "         source=empty         dummy key required for asg driver.",
  " ",
  " ------------------------------------------------------------------------",
  " Trace info:",
  " ",
  "          data  = <path>    data file, SU format - must exist ",
  "                            headers establish acquisition geometry",
  "                            output for forward modeling (including all",
  "                            orders of derivative), input for adjoint",
  " ",
  " ------------------------------------------------------------------------",
  " Model info:",
  " ",
  "        csq = <path>      input velocity squared model, ",
  "                          determines simulation spatial grid,",
  " ",
  " ------------------------------------------------------------------------",
  " MPI info:",
  " ",
  "        mpi_np1 = 1           number of subdomains along axis 1",
  "        mpi_np2 = 1           number of subdomains along axis 2",
  "        mpi_np3 = 1           number of subdomains along axis 3",
  "        partask = 1           number of shots to execute in parallel",
  " ",
  " ------------------------------------------------------------------------",
  " Output info:",
  " ",
  " FD ouput - written to coutxxx.txt on rank xxx",
  "       printact = 0           monitor output:",
  "                            < 0 - none",
  "                              0 - announce each shot simulation",
  "                              1 - print time step index",
  "                              2 - diagnostic messages from main routines",
  "                              > 2 - much more, useful only for serious ",
  "                                 debugging",
  "        dump_pi = 0           dump parallel/dom. decomp info",
  "       dump_lda = 0           dump grid data for allocated arrays",
  "       dump_ldc = 0           dump grid data for computational arrays",
  "       dump_ldr = 0           dump grid data for receive buffers",
  "       dump_lds = 0           dump grid data for send buffers",
  "      dump_term = 0           dump trace header data",
  " ",
  " ------------------------------------------------------------------------",
  " MPS info:",
  " ",
  "        appx_ord = 4       singular source approximation order,",
  "                           should match spatial order of FD method, i.e., 2*order",
  "       MPS_ord_0 = 0       MPS order in z-direction, s_0",
  "       MPS_ord_1 = 0       MPS order in x-direction, s_1",
  "       MPS_ord_2 = 0       MPS order in y-direction, s_2",
  "        MPS_file = <path>  MPS file containing MPS coefficients, SU format - must exist",
  "                           Must contain source locations.",
  " ---------------------------end parameters ------------------------------",
  NULL };


IOKEY IWaveInfo::iwave_iokeys[]
= {
  {"csq",      0, true,  1 },
  {"data",     1, false, 2 },
  {"source",   1, true,  2 },
  {"",         0, false, 0 }
};


int xargc;
char **xargv;

int main(int argc, char ** argv) {

  try {

#ifdef IWAVE_USE_MPI
    int ts=0;
    MPI_Init_thread(&argc,&argv,MPI_THREAD_FUNNELED,&ts);        
#endif

    PARARRAY * pars = NULL;
    FILE * stream = NULL;
    TSOpt::IWaveEnvironment(argc, argv, 0, &pars, &stream);
    ps_printall(*pars,stderr);

#ifdef IWAVE_USE_MPI
    if(retrieveGlobalRank()==0){
#endif

      if(argc<2){
	pagedoc();
	exit(0);
      }

      cerr << "\n////////////////////////////////////////////////////////////\n"
	   << "Running acd_scal.x\n";

#ifdef IWAVE_USE_MPI
    }
#endif

    TSOpt::MPS_KEYS mks;
    mks.appx_ord = "appx_ord";
    mks.MPS_file = "MPS_file";
    mks.grid_file = "csq";
    mks.RHS_files.push_back("source");

    //initializing MPS space
    Scal_MPS_Space mps_sp(mks,*pars);
    
    //initializing MPS_to_RHS
    MPS_to_RHS m2r(mks,*pars,mps_sp);
    
    //constructing MPS IWaveLOVOp
    TSOpt::IWaveLOVOp lovop(m2r.get_pars(),stream);
    RVL::LinCompLOVOp<float> mps_lovop(m2r,lovop);

    //setting up domain vector
    Vector<ireal> x(mps_lovop.getDomain());
    Components<ireal> cx(x);

    Components<ireal> cx_m(cx[0]); //model params
    RVL::AssignFilename af_csq(valparse<string>(*pars,"csq"));
    cx_m[0].eval(af_csq);
    Components<ireal> cx_f(cx[1]); //source params
    RVL::AssignFilename af_mps(valparse<string>(*pars,"MPS_file"));
    cx_f[0].eval(af_mps);

    //setting up range vector
    Vector<ireal> y(mps_lovop.getRange());
    RVL::AssignFilename af_data(RVL::valparse<string>(*pars,"data"));
    y.eval(af_data);
    
    //applying op
    RVL::OperatorEvaluation<ireal> eval(mps_lovop,x);
    y.copy(eval.getValue());

    //clean up
    ps_delete(&pars);
    iwave_fdestroy();


#ifdef IWAVE_USE_MPI
    if(retrieveGlobalRank()==0){
#endif

      cerr << "\nFinishing acd_scal.x\n"
	   << "////////////////////////////////////////////////////////////\n";

#ifdef IWAVE_USE_MPI
    }
    MPI_Finalize();
#endif

  }
  catch(RVLException &e){
    e << "Exiting with error!\n";
#ifdef IWAVE_USE_MPI
    if( retrieveGlobalRank()==0 ){
#endif
      e.write(cerr);
#ifdef IWAVE_USE_MPI
    }
    MPI_Barrier(retrieveGlobalComm());
    MPI_Abort(retrieveGlobalComm(),0);
#endif 
    exit(1);
  }
}
